@@ -35,6 +35,60 @@ namespace {
 
         return stack_trace;
     }
+
+    const string
+        sym_ARGV    = "ARGV",
+        sym_ARGVOUT = "ARGVOUT",
+        sym_ENV     = "ENV",
+        sym_INC     = "INC",
+        sym_SIG     = "SIG",
+        sym_STDIN   = "STDIN",
+        sym_STDOUT  = "STDOUT",
+        sym_STDERR  = "STDERR",
+
+        sym_BEGIN     = "BEGIN",
+        sym_CHECK     = "CHECK",
+        sym_END       = "END",
+        sym_INIT      = "INIT",
+        sym_UNITCHECK = "UNITCHECK";
+
+    // does the same job as S_gv_is_in_main in gv.c
+    bool is_in_main(const string &name) {
+        size_t length = name.length();
+
+        switch (name[0]) {
+        case '_':
+            return length == 1;
+        case 'A':
+            return name == sym_ARGV || name == sym_ARGVOUT;
+        case 'E':
+            return name == sym_ENV;
+        case 'I':
+            return name == sym_INC;
+        case 'S':
+            return name == sym_SIG || name == sym_STDIN ||
+                   name == sym_STDOUT || name == sym_STDERR;
+        default:
+            return false;
+        }
+    }
+
+    bool is_special_sub(const string &name) {
+        switch (name[0]) {
+        case 'B':
+            return name == sym_BEGIN;
+        case 'C':
+            return name == sym_CHECK;
+        case 'E':
+            return name == sym_END;
+        case 'I':
+            return name == sym_INIT;
+        case 'U':
+            return name == sym_UNITCHECK;
+        default:
+            return false;
+        }
+    }
 }
 
 MappingOptions::MappingOptions(pTHX_ SV *options_ref) :
@@ -185,6 +239,14 @@ namespace {
         return 0;
     }
 
+    int dup_refcounted(pTHX_ MAGIC *mg, CLONE_PARAMS *param) {
+        Refcounted *refcounted = (Refcounted *) mg->mg_ptr;
+
+        refcounted->ref();
+
+        return 0;
+    }
+
     MGVTBL manage_refcounted = {
         NULL, // get
         NULL, // set
@@ -192,10 +254,23 @@ namespace {
         NULL, // clear
         free_refcounted,
         NULL, // copy
-        NULL, // dup
+        dup_refcounted,
         NULL, // local
     };
 
+    void bind_special_constant(pTHX_ const string &target, const string &perl_package, UV value) {
+        static const char xsubname[] = "Google::ProtocolBuffers::Dynamic::Mapper::constant";
+
+        CV *src = get_cv(xsubname, 0);
+        CV *new_xs = newXS((perl_package + "::" + target + "*").c_str(), CvXSUB(src), __FILE__);
+        CvXSUBANY(new_xs).any_uv = value;
+
+        GV *gv = gv_fetchpv((perl_package + "::" + target).c_str(),
+                            GV_ADD, SVt_PVCV);
+
+        GvCV_set(gv, new_xs);
+    }
+
     void copy_and_bind(pTHX_ const char *name, const char *target, const string &perl_package, Refcounted *refcounted, void *obj) {
         static const char prefix[] = "Google::ProtocolBuffers::Dynamic::Mapper::";
         size_t length = strlen(name);
@@ -208,9 +283,10 @@ namespace {
         CV *new_xs = newXS((perl_package + "::" + target).c_str(), CvXSUB(src), __FILE__);
 
         CvXSUBANY(new_xs).any_ptr = obj;
-        sv_magicext((SV *) new_xs, NULL,
+        MAGIC *mg = sv_magicext((SV *) new_xs, NULL,
                     PERL_MAGIC_ext, &manage_refcounted,
                     (const char *) refcounted, 0);
+        mg->mg_flags |= MGf_DUP;
         refcounted->ref();
     }
 
@@ -653,8 +729,17 @@ void Dynamic::bind_enum(pTHX_ const string &perl_package, const EnumDescriptor *
     for (int i = 0, max = descriptor->value_count(); i < max; ++i) {
         const EnumValueDescriptor *value = descriptor->value(i);
 
-        newCONSTSUB(stash, value->name().c_str(),
-                    newSVuv(value->number()));
+        // due to an implementation quirk, some special names such as STDERR
+        // are not looked up in stash, and need to be fully qualified
+        if (is_in_main(value->name())) {
+            newCONSTSUB(stash, (perl_package + "::" + value->name()).c_str(),
+                        newSVuv(value->number()));
+        } else if (is_special_sub(value->name())) {
+            bind_special_constant(aTHX_ value->name(), perl_package, value->number());
+        } else {
+            newCONSTSUB(stash, value->name().c_str(),
+                        newSVuv(value->number()));
+        }
     }
 }
 
